from dataclasses import dataclass
import ffmpeg
import os
import logging as log

IMAGE_DIR = os.getenv('IMAGE_DIR') or './images'


@dataclass
class VideoBase:

    video_name: str = "output"
    video_bitrate: str = "5000k"

    @property
    def image_directory(self):
        return f'{IMAGE_DIR}/{self.video_name}'

    @property
    def output_file(self):
        return f'{self.image_directory}/{self.video_name}.mp4'

    @property
    def input_glob(self):
        return f'{self.image_directory}/*.jpg'

    def input_directory_exists(self):
        log.info(f'[FFMPEG] Checking for input directory: {self.image_directory}')
        return False if not os.path.exists(self.image_directory) else True

    def input_directory_not_empty(self):
        log.info(f'[FFMPEG] Checking for images in input directory: {self.image_directory} - file count: {len(os.listdir(self.image_directory))}')
        return False if len(os.listdir(self.image_directory)) == 0 else True


'''
Youtube Shorts video ffmpeg command example:
ffmpeg -framerate 15 -pattern_type glob -i "*_GOES16-*-GEOCOLOR-*.jpg" -i space_walk_short.mp3 -map 0:v -map 1:a -c:v libx264 -c:a aac
  -b:a 192k -r 24 -vf "scale=w=1080:h=1920,setsar=1" -af "afade=t=out:st=280:d=3" output.mp4

'''


@dataclass
class ShortsVideoManager(VideoBase):

    framerate: int = 15
    audiofile: str = "./audio/space_walk_short.mp3"
    vcodec: str = "libx264"
    acodec: str = "aac"
    audio_bitrate: str = "192k"

    def build(self):
        log.info(f'[FFMPEG] Creating shorts video for job: {self.video_name}')

        input_video = ffmpeg.input(self.input_glob, pattern_type='glob', framerate=self.framerate).filter("scale", 1080, 1920).filter("setsar", 1)
        input_audio = ffmpeg.input(self.audiofile)

        (
            ffmpeg
            .output(input_audio,
                    input_video, self.output_file,
                    vcodec=self.vcodec,
                    acodec=self.acodec,
                    audio_bitrate=self.audio_bitrate,
                    video_bitrate=self.video_bitrate,
                    shortest=None)
            .run()
        )


'''
Youtube video ffmpeg command example:
ffmpeg -framerate 15 -pattern_type glob -i *_GOES16-*-GEOCOLOR-*.jpg -c:v libx264 -r 24 -vf scale=1920x1080 ./output.mp4
'''


@dataclass
class VideoManager(VideoBase):

    framerate: int = 15
    vcodec: str = "libx264"

    def build(self):
        log.info(f'[FFMPEG] Creating video for job: {self.video_name}')
        (
            ffmpeg
            .input(self.input_glob, pattern_type='glob', framerate=self.framerate)
            .filter("scale", 1920, 1080)
            .output(self.output_file, vcodec=self.vcodec, video_bitrate=self.video_bitrate)
            .run()
        )



from dataclasses import dataclass


@dataclass
class RegionURLManager:
    alaska: str = "https://cdn.star.nesdis.noaa.gov/GOES18/ABI/SECTOR/ak/GEOCOLOR/"
    central_alaska: str = "https://cdn.star.nesdis.noaa.gov/GOES18/ABI/SECTOR/cak/GEOCOLOR/"
    south_mississippi_valley: str = "https://cdn.star.nesdis.noaa.gov/GOES16/ABI/SECTOR/smv/GEOCOLOR/"
    north_mississippi_valley: str = "https://cdn.star.nesdis.noaa.gov/GOES16/ABI/SECTOR/umv/GEOCOLOR/"
    pac_west: str = "https://cdn.star.nesdis.noaa.gov/GOES18/ABI/SECTOR/wus/GEOCOLOR/"
    mexico: str = "https://cdn.star.nesdis.noaa.gov/GOES16/ABI/SECTOR/mex/GEOCOLOR/"
    conus: str = "https://cdn.star.nesdis.noaa.gov/GOES16/ABI/CONUS/GEOCOLOR/"

    def __getitem__(self, item):
        return getattr(self, item)

    def valid_region(self, region: str):
        return region in self.__dict__.keys()



from fastapi import FastAPI
from ManagerAPI.api.routers import jobs, auth

managerAPI = FastAPI()


@managerAPI.get('/health')
async def health():
    return {'managerAPI': 'up'}

'''
Include additional routers
'''
managerAPI.include_router(jobs.jobRouter)
managerAPI.include_router(auth.authRouter)



from fastapi.testclient import TestClient
from ManagerAPI import managerAPI

client = TestClient(managerAPI)


def test_root():
    response = client.get("/health")
    assert response.status_code == 200
    assert response.json() == {"managerAPI": "up"}



from uuid import uuid4
from datetime import datetime
from time import time
from pydantic import BaseModel, Field
from typing import Union, Optional


class ManagedJobModel(BaseModel):
    job_id: Optional[str] = Field(default_factory=lambda: str(uuid4()))
    job_status: Union[str, None] = "new"
    job_start_time: Optional[int] = Field(default_factory=lambda: int(time()))
    job_end_time: Union[int, None] = -1
    job_error: Union[str, None] = ''
    img_date: str
    image_links: Union[str, None] = ''
    gif_urn: Union[str, None] = ''
    video_urn: Union[str, None] = ''
    img_resolution: str
    yt_video_id: Union[str, None] = ''
    region: str

    class Config:
        orm_mode = True

    def get_formatted_date(self):
        date_str = self.img_date.split("-")
        return datetime(int(date_str[0]), int(date_str[1]), int(date_str[2])).strftime("%Y%j")



from LogHandler.logHandler import LogHandler
from JobHandler.jobHandler import JobHandler
from ManagerAPI.api.models.managedJob import ManagedJobModel
from VideoMaker.videoMaker.managers.video_manager import VideoBase, VideoManager, ShortsVideoManager

from time import sleep
import logging as log


def create_video(job: ManagedJobModel):

    log.info(f'Creating video for job: {job.job_id}')
    job_resolution = job.img_resolution.split("x")

    if not VideoBase(video_name=str(job.job_id)).input_directory_exists() or not VideoBase(video_name=str(job.job_id)).input_directory_exists():
        log.error(f'No images found for job: {job.job_id}')
        job.job_status = "error"
        job.job_error += f'No images found for job: {job.job_id} ,'
        return

    log.debug(f'Job resolution: {job_resolution[0]}x{job_resolution[1]}')
    if job_resolution[0] != job_resolution[1]:
        VideoManager(video_name=str(job.job_id)).build()
    else:
        ShortsVideoManager(video_name=str(job.job_id)).build()

    job.job_status = "packed"
    job.video_urn = VideoBase(video_name=str(job.job_id)).output_file

    return


def run():
    jobHandler = JobHandler("grabbed")
    log.info('Checking for grabbed jobs')

    if not jobHandler.available_job():
        log.debug('No grabbed jobs found')
        return

    new_job = jobHandler.fetch_job()

    if not new_job:
        log.error(f'Failed to fetch job {new_job.job_id}')
        return

    create_video(new_job)

    if new_job.job_status == "error":
        jobHandler.update_job(new_job)
        log.error(f'Updated job {new_job.job_id} with error')
        return

    jobHandler.update_job(new_job)

    return


def launch():

    LogHandler(service_name="videoMaker").bootstrap()
    log.info("Starting VideoMaker service")

    while True:
        sleep(5)
        run()


if __name__ == "__main__":

    launch()



from fastapi import APIRouter, Request
from fastapi.responses import RedirectResponse
from ManagerAPI.api.managers.connection_manager import AuthManager, RedisConnectionConfig
import logging
import google_auth_oauthlib.flow
import os


log = logging.getLogger('uvicorn')
authRouter = APIRouter()

session = {}

# Google API credentials
CLIENT_SECRETS_FILE = os.getenv('CLIENT_SECRETS_FILE') or 'client_secrets.json'
FLOW_BASE_URI = os.getenv('FLOW_BASE_URI') or 'localhost'
SCOPES = ['https://www.googleapis.com/auth/youtube.upload']


@authRouter.get("/auth")
async def auth_init():

    log.info('Initializing auth')
    """Initialize auth and redirect"""
    flow = google_auth_oauthlib.flow.Flow.from_client_secrets_file(
        CLIENT_SECRETS_FILE,
        scopes=SCOPES
    )

    flow.redirect_uri = f'https://{FLOW_BASE_URI}/callback'

    log.info(f'Fetching authorization url - {flow.redirect_uri}')
    authorization_url, state = flow.authorization_url(
        access_type='offline',
        include_granted_scopes='true'
    )

    session['state'] = state

    return RedirectResponse(authorization_url)


@authRouter.get("/callback")
async def auth_callback(request: Request):

    log.info('[Callback] Auth')
    """Verify login"""
    state = session['state']

    flow = google_auth_oauthlib.flow.Flow.from_client_secrets_file(
        CLIENT_SECRETS_FILE,
        scopes=SCOPES,
        state=state
    )

    flow.redirect_uri = f'https://{FLOW_BASE_URI}/callback'
    authorization_response = f'https://{FLOW_BASE_URI}{request.url.path}?{request.url.query}'
    log.debug(f'[Callback] Authorization Response: {authorization_response}')

    flow.fetch_token(authorization_response=authorization_response)
    credentials = flow.credentials

    return AuthManager(refresh_token=credentials.refresh_token, connection_config=RedisConnectionConfig()).store()


@authRouter.get("/refresh_token")
async def refresh_token():

    log.info('Refreshing token')

    return AuthManager(connection_config=RedisConnectionConfig()).fetch()



from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from googleapiclient.http import MediaFileUpload
import google.oauth2.credentials
from dataclasses import dataclass
import os
import json
import requests
import logging as log


@dataclass
class YTAuthManager:
    secrets_file: str = os.getenv("CLIENT_SECRETS_FILE", "client_secrets.json")
    client_id: str = ""
    client_secret: str = ""
    token_url: str = "https://oauth2.googleapis.com/token"
    access_token: str = ""
    refresh_token: str = ""
    manager_url: str = os.getenv("MANAGER_URL", "http://localhost")
    youtube_api_service_name: str = "youtube"
    youtube_api_version: str = "v3"

    def fetch_refresh_token(self):

        try:
            log.debug(f'[VideoAuthManager] fetching refresh token from {self.manager_url}/refresh_token')
            refresh_token = requests.get(f'http://{self.manager_url}/refresh_token').json()
        except requests.exceptions.ConnectionError:
            return False
        except requests.exceptions.ReadTimeout:
            return False

        log.debug(f'[VideoAuthManager] fetched refresh token {refresh_token}')

        self.refresh_token = refresh_token['refresh_token']

        return True

    def pull_secrets(self):

        with open(self.secrets_file) as f:
            secrets = json.load(f)

        self.client_id = secrets["web"]["client_id"]
        self.client_secret = secrets["web"]["client_secret"]
        self.token_url = secrets["web"]["token_uri"]

        log.debug(f'[VideoAuthManager] pulled secrets {self.client_id} {self.client_secret} {self.token_url}')
        return True

    def fetch_credentials(self):

        params = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "refresh_token": self.refresh_token,
            "grant_type": "refresh_token"
        }

        try:
            access_token_response = requests.post(self.token_url, params=params)
        except requests.exceptions.ConnectionError:
            return False

        log.debug(f'[VideoAuthManager] fetched credentials {access_token_response.json()}')

        return google.oauth2.credentials.Credentials(access_token_response.json()['access_token'])

    def bootstrap(self):

        if not self.client_id:
            self.pull_secrets()

        if not self.refresh_token:
            self.fetch_refresh_token()

        return self


@dataclass
class YTVideoManager:

    title: str = ""
    description: str = ""
    video_type: str = ""
    privacy_status: str = "private"
    video_file: str = ""
    auth_manager: YTAuthManager = None

    def upload_video(self, credentials):

        if self.video_type == "shorts":
            self.description += " #Shorts"

        try:
            youtube = build(
                self.auth_manager.youtube_api_service_name,
                self.auth_manager.youtube_api_version,
                credentials=credentials
            )

            video = youtube.videos().insert(
                part="snippet,status",
                body={
                      "snippet": {
                         "title": self.title,
                         "description": self.description,
                      },
                      "status": {
                         "privacyStatus": self.privacy_status
                      }
                },
                media_body=MediaFileUpload(self.video_file, chunksize=-1, resumable=True)
            ).execute()

        except HttpError as e:
            log.error(f'Video Upload error - {e.resp.status} occurred: {e.content}')
            return False

        return video['id'] if "id" in video else False



from setuptools import setup, find_packages

setup(
    name='RootsWeatherProject',
    version='1.0.6',
    packages=find_packages(),
    install_requires=['pytest', 'fastapi', 'pydantic', 'uvicorn', 'httpx', 'redis', 'requests',
                      'beautifulsoup4', 'bs4', 'setuptools', 'ffmpeg-python', 'google_auth_oauthlib',
                      'google-api-python-client'],
    author='John Anthony',
    description='A package holding microservices to create video from compiled weather image data',
    url='https://github.com/johncanthony/RootsWeatherProject',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.12'
    ],
    entry_points={
        'console_scripts': [
            'managerAPI = ManagerAPI.server:launch',
            'imageResolver = ImageResolver.imageResolver.resolver:launch',
            'imageGrabber = ImageGrabber.imageGrabber.grabber:launch',
            'videoMaker = VideoMaker.videoMaker.videomaker:launch',
            'videoUploader = VideoUploader.videoUploader.uploader:launch'
        ]
    }
)



import pytest
from unittest import mock
from JobHandler.jobHandler import JobHandler
from ManagerAPI.api.models.managedJob import ManagedJobModel


@pytest.fixture
def job_handler():
    return JobHandler("new")


def test_available_job(job_handler):
    # Mock the external request
    with mock.patch('requests.get') as mock_get:
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = {'jobs:NEW': ['b40df782-9c48-4159-bcc0-fc5cd55e0c82'], 'jobs:IMGRESOLVED': ['b40df782-9c48-4159-bcc0-fc5cd55e0c83'],
                                                   'jobs:GRABBED': ['b40df782-9c48-4159-bcc0-fc5cd55e0c84'], 'jobs:PACKED': ['b40df782-9c48-4159-bcc0-fc5cd55e0c85'],
                                                   'jobs:UPLOADED': ['b40df782-9c48-4159-bcc0-fc5cd55e0c86'], 'jobs:ERROR': ['b40df782-9c48-4159-bcc0-fc5cd55e0c87']}

        # Call the method under test
        status = job_handler.available_job()

        # Assert the expected result
        assert status is True
        mock_get.assert_called_once_with('http://localhost:8000/jobs')


def test_fetch_job_fields(job_handler):
    # Mock the external request
    with mock.patch('requests.get') as mock_post:
        mock_post.return_value.status_code = 200
        mock_post.return_value.json.return_value = {
            "new": {
                "job_id": "efb7f53b-e5a1-4c30-9f45-4b6c116a36ca",
                "job_status": "new",
                "job_start_time": 1705313815,
                "job_end_time": -1,
                "job_error": "",
                "img_date": "2024-01-12",
                "image_links": "",
                "gif_urn": "",
                "video_urn": "",
                "img_resolution": "1250x750",
                "yt_video_id": "",
                "region": "CONUS"
                }
            }

        # Call the method under test
        job_id = job_handler.fetch_job()
        assert type(job_id) is ManagedJobModel

        # Assert the expected result
        mock_post.assert_called_once_with('http://localhost:8000/job/queue/new')


def test_update_job(job_handler):

    with mock.patch('requests.post') as mock_post:
        mock_post.return_value.status_code = 200
        mock_post.return_value.json.return_value = {
            "new": {
                "job_id": "efb7f53b-e5a1-4c30-9f45-4b6c116a36ca",
                "job_status": "new",
                "job_start_time": 1705313815,
                "job_end_time": -1,
                "job_error": "",
                "img_date": "2024-01-12",
                "image_links": "",
                "gif_urn": "",
                "video_urn": "",
                "img_resolution": "1250x750",
                "yt_video_id": "",
                "region": "CONUS"
                }
            }

        mock_job = ManagedJobModel(job_id="efb7f53b-e5a1-4c30-9f45-4b6c116a36ca", job_status="new", job_start_time=1705313815,
                                   job_end_time=-1, job_error="", img_date="2024-01-12", image_links="", gif_urn="", video_urn="",
                                   img_resolution="1250x750", yt_video_id="", region="CONUS")
        job_push = job_handler.update_job(mock_job)

        assert job_push is True
        mock_post.assert_called_once_with('http://localhost:8000/job', json=mock_job.model_dump())



from LogHandler.logHandler import LogHandler
import logging as log
from logging.handlers import TimedRotatingFileHandler
import unittest


class TestLogHandler(unittest.TestCase):
    def test_logHandler(self):
        # Create an instance of LogHandler

        log_handler = LogHandler(service_name='test', level='DEBUG')
        log_handler.bootstrap()

        # Assert Default getenv() values are set
        self.assertEqual(log_handler.service_name, 'test')
        self.assertEqual(log_handler.dir, './')

        # Assert the appropriate log level and log file name are set
        self.assertEqual(log.getLogger().level, log.DEBUG)
        testLogHandler = [handler.baseFilename for handler in log.getLogger().handlers if isinstance(handler, TimedRotatingFileHandler)]
        self.assertIn('test-service.log', testLogHandler[0])


if __name__ == '__main__':
    unittest.main()
